/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.listeners;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState;
import net.yale.cs.p4p.p2p.azureus.client.data.Torrents;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState.PeerRecord;
import net.yale.cs.p4p.p2p.azureus.client.util.StringUtil;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.peers.PeerManagerListener;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class P4PPeerManagerListener implements PeerManagerListener {

	/* Unchoking temporarily disabled */
	
//	private static final int UNCHOKE_SCHEDULE_INTERVAL = 5 * 1000;
//	
//	private class UnchokeTask extends TimerTask {
//
//		/* Keep track of last peers we saw that were unchoked.  We will
//		 * use this to tell when there is a new round for which we will
//		 * select an optimistically-unchoked peer. */
//		private Set<String> lastUnchoked_ = new HashSet<String>();
//		
//		public boolean isNewRound(Peer[] peers, Set<String> unchokedPeers) {
//			boolean result = false;
//			
//			int unchokeCount = 0;
//			for (Peer peer : peers) {
//				if (peer.isChoking())
//					continue;
//
//				String peerString = peer.getIp() + ':' + peer.getPort();
//				unchokedPeers.add(peerString);
//
//				++unchokeCount;
//				if (!lastUnchoked_.contains(peerString))
//					result = true;
//			}
//			if (unchokeCount != lastUnchoked_.size())
//				result = true;
//			
//			return result;
//		}
//
//		public void run() {
//			
//			Peer[] peers = manager_.getPeers();
//			if (peers == null)
//				return;
//
//			Download download;
//			try {
//				download = manager_.getDownload();
//			} catch (DownloadException e) {
//				plugin_.getLogger().log("Failed to schedule new optimistic unchokes: " + e);
//				return;
//			}
//			
//			/* Ignore if P4P is not enabled */
//			if (!plugin_.getExperiment().isP4PEnabled(download))
//				return;
//
//			/* Ignore if no torrent */
//			Torrent torrent = download.getTorrent();
//			if (torrent == null)
//				return;
//
//			Set<String> unchokedPeers = new HashSet<String>();
//
//			/* Skip if there were no changes. That is, Vuze has not yet
//			 * redone its unchokes */
//			if (!isNewRound(peers, unchokedPeers))
//				return;
//
//			plugin_.getLogger().log(torrent, "Scheduling new optimistic unchoke");
//			plugin_.getPeerMonitor().updateOptimisticUnchoke(download);
//			
//			lastUnchoked_ = unchokedPeers;
//		}
//	}
	
	private P4PPlugin plugin_;
	private Torrents torrents_;
	private PeerManager manager_;
	
//	private Timer unchokeTimer_ = new Timer();

	public P4PPeerManagerListener(P4PPlugin plugin, PeerManager manager) {
		plugin_ = plugin;
		torrents_ = plugin_.getTorrents();
		manager_ = manager;
//		unchokeTimer_.schedule(new UnchokeTask(), 0, UNCHOKE_SCHEDULE_INTERVAL);
	}

	public void peerAdded(PeerManager manager, Peer peer) {
		try {
			Download download = manager_.getDownload();
			if (download == null)
				return;
			
			Torrent torrent = download.getTorrent();
			if (torrent == null)
				return;

			// plugin_.getLogger().log(torrent, "Connected to peer " + peer.getIp());

			TorrentState torrentState = torrents_.getTorrent(torrent);

			/* Mark peers in Vuze-biased torrents as such */
			if (torrentState.getVuzeBiased())
				peer.setUserData("LOG_TAGS", "alg=VUZEBIAS");

			/* Notify torrent state that new peer is connected */
			PeerRecord rec = torrentState.peerConnected(peer);
			if (rec != null) {
				/* Set our logging properties for this peer. These are read by the
				 * Instrumentation plugin. */
				peer.setUserData("LOG_KIND", rec.getPeerSource());
				peer.setUserData("LOG_TAGS", StringUtil.join(rec.getSchedInfo().getLogTags(), ','));
			}

			/* Exit if we're biased or aren't P4P enabled */
			if (torrentState.getVuzeBiased() || !plugin_.getExperiment().isP4PEnabled(download))
				return;
	
			/* Send notification to check if we need more peers */
			plugin_.getPeerMonitor().reschedule(download, false);

		} catch (DownloadException e) {
			plugin_.getLogger().log("Failed to handle peer arrival: " + e);
		}
	}

	public void peerRemoved(PeerManager manager, Peer peer) {
		try {
			Download download = manager_.getDownload();
			if (download == null)
				return;
			
			Torrent torrent = download.getTorrent();
			if (torrent == null)
				return;
			
			TorrentState torrentState = torrents_.getTorrent(torrent);
			
			// plugin_.getLogger().log(torrent, "Disconnected from peer " + peer.getIp());

			/* Notify torrent state that peer is disconnected */
			torrentState.peerDisconnected(peer);

			/* Exit if we're biased or are aren't P4P enabled */
			if (torrentState.getVuzeBiased() || !plugin_.getExperiment().isP4PEnabled(download))
				return;

			/* Send notification to check if we need more peers */
			plugin_.getPeerMonitor().reschedule(download, false);

		} catch (DownloadException e) {
			plugin_.getLogger().log("Failed to handle peer removal: " + e);
		}
	}

	public void cleanup() {
//		unchokeTimer_.cancel();
	}
}
